package windsock;

import windsock.*;
import java.net.*;
import java.io.*;

public class SocksSocket extends Socket {
    public static final byte METHOD_NOAUTH = 0x00;
    public static final byte METHOD_GSSAPI = 0x01;
    public static final byte METHOD_USERPASS = 0x02;
    public static final byte METHOD_USER = (byte) 0x80;
    public static final byte METHOD_NONE = (byte) 0xff;

    public static final byte CMD_CONNECT = 0x01;
    public static final byte CMD_BIND = 0x02;
    public static final byte CMD_UDP_ASSOC = 0x03;

    public static final byte ATYP_IPV4 = 0x01;
    public static final byte ATYP_DOMAIN = 0x03;
    public static final byte ATYP_IPV6 = 0x04;

    public static final byte REPLY_SUCCESS = 0x00;
    public static final byte REPLY_FAIL = 0x01;
    public static final byte REPLY_DISALLOWED = 0x02;
    public static final byte REPLY_NET_UNR = 0x03;
    public static final byte REPLY_HOST_UNR = 0x04;
    public static final byte REPLY_REFUSED = 0x05;
    public static final byte REPLY_TTL_EXP = 0x06;
    public static final byte REPLY_CMD_UNSUP = 0x07;
    public static final byte REPLY_ATYP_UNSUP = 0x08;

    public static final byte preferredVersion = 5;

    private byte version = 0;
    private Socket sock = null;
    private InputStream in = null;
    private OutputStream out = null;
    private BufferedOutputStream bout = null;

    private Request v4Request = null;
    private String v4User = null;

    public InputStream getInputStream() throws IOException {
	return in;
    }

    public OutputStream getOutputStream() throws IOException {
	bout.flush();
	return out;
    }

    public SocksSocket(Socket sock) throws IOException {
	this.sock = sock;
	in = sock.getInputStream();
	out = sock.getOutputStream();
	bout = new BufferedOutputStream(out);
    }

    private void checkVersion() throws SocksException, IOException {
	int version = in.read();

	if (version == -1) {
	    throw new SocksException("Premature EOF");
	}

	if (this.version == 0) {
	    if (version < 4 || version > 5) {
		throw new SocksException("Unsupported version: " + version);
	    }
	    this.version = (byte) version;
	} else if (this.version != version) {
	    throw new SocksException("Using version " + this.version +
		    " but got version " + version);
	}
    }

    private void setVersion() {
	if (version == -1) {
	    version = preferredVersion;
	}
    }

    public byte getMethod(byte[] supported)
	    throws SocksException, IOException {
	checkVersion();

	byte[] methods = null;

	if (version == 4) {
	    methods = new byte[2];
	    methods[0] = METHOD_NOAUTH;
	    methods[1] = METHOD_USER;

	    // read the whole request frame now, because we need to read the
	    // user at the end.

	    v4Request = new Request();

	    int i = in.read();
	    if (i == -1) {
		throw new IOException("Premature EOF");
	    }
	    // command codes are the same in socks4/5
	    v4Request.command = (byte) i;

	    v4Request.port = in.read() * 256;
	    if (v4Request.port < 0) {
		throw new IOException("Premature EOF");
	    }
	    v4Request.port += (i = in.read());
	    if (i < 0) {
		throw new IOException("Premature EOF");
	    }

	    v4Request.addressType = ATYP_IPV4;
	    v4Request.address = new byte[4];
	    read(v4Request.address);

	    v4User = "";
	    while ((i = in.read()) > 0) {
		v4User += (char) i;
	    }
	    if (i < 0) {
		throw new IOException("Premature EOF");
	    }

	} else if (version == 5) {
	    int nmethods = in.read();
	    if (nmethods < 1) {
		throw new IOException("Read invalid nmethods from " + in +
			": " + nmethods);
	    }

	    int offset = 0;
	    methods = new byte[nmethods];
	    read(methods);
	}

	// now check which method to actually use
	for (int sup = 0; sup < supported.length; sup++) {
	    for (int meth = 0; meth < methods.length; meth++) {
		if (methods[meth] == supported[sup]) {
		    if (version == 5) {
			bout.write(5);
			bout.write(methods[meth]);
			bout.flush();
		    }
		    return methods[meth];
		}
	    }
	}

	throw new SocksException("No supported method found.");
    }

    public Request getRequest() throws IOException, SocksException {
	if (version != 4)
	    checkVersion();

	Request r = new Request();
	int i;

	if (version == 4) {
	    return v4Request;

	} else if (version == 5) {
	    r.command = (byte) (i = in.read());
	    if (i == -1) {
		throw new SocksException("Premature EOF");
	    }
	    if (r.command != CMD_CONNECT &&
		    r.command != CMD_BIND &&
		    r.command != CMD_UDP_ASSOC) {
		throw new IOException("Read invalid command from " + in + ": " +
			r.command);
	    }

	    byte reserved = (byte) (i = in.read());
	    if (i == -1) {
		throw new SocksException("Premature EOF");
	    }
	    if (reserved != 0) {
		throw new IOException("Reserved byte not set to 0 from " + in);
	    }

	    r.addressType = (byte) (i = in.read());
	    if (i == -1) {
		throw new SocksException("Premature EOF");
	    }
	    switch (r.addressType) {
		case ATYP_IPV4:
		    r.address = new byte[4];
		    read(r.address);
		    break;

		case ATYP_DOMAIN:
		    int size = in.read();
		    if (size < 1) {
			return null;
		    }
		    r.address = new byte[size];
		    read(r.address);
		    break;

		case ATYP_IPV6:
		    r.address = new byte[16];
		    read(r.address);
		    break;

		default:
		    throw new SocksException("Unknown address type : " +
			    r.addressType);
	    }

	    r.port = in.read() * 256;
	    if (r.port < 0) {
		throw new IOException("Premature EOF");
	    }
	    r.port += (i = in.read());
	    if (i < 0) {
		throw new IOException("Premature EOF");
	    }
	}

	return r;
    }

    public void sendReply(Request r, byte reply)
	    throws IOException, SocksException {
	setVersion();

	Log.log("sendReply(" + r + ", " + reply + ")");

	if (version == 4) {
	    bout.write(4);
	    if (reply == REPLY_SUCCESS) {
		bout.write(90);
	    } else {
		bout.write(91);
	    }
	    bout.write(r.port >> 8);
	    bout.write(r.port & 0xff);
	    if (r.addressType != ATYP_IPV4) {
		throw new SocksException("Invalid address type for SOCKS4");
	    }
	    bout.write(r.address);
	    bout.flush();

	} else if (version == 5) {
	    bout.write(5);
	    bout.write(reply);
	    bout.write(0);
	    bout.write(r.addressType);
	    if (r.addressType == ATYP_DOMAIN) {
		bout.write(r.address.length);
	    }
	    bout.write(r.address);
	    bout.write(r.port >> 8);
	    bout.write(r.port & 0xff);
	    bout.flush();
	}
    }

    protected void read(byte[] buf) throws IOException, SocksException {
	int offset = 0;
	while (offset < buf.length) {
	    int ret = in.read(buf, offset, buf.length - offset);
	    if (ret == -1) {
		throw new SocksException("Premature EOF");
	    }
	    offset += ret;
	}
    }
}
